import math
from itertools import repeat
from typing import Type, Any, Callable, Union, List, Optional

import torch
from torch import Tensor
import torch.nn as nn
import torch.nn.functional as F

from .sri_conv import SRI_Conv2d

def _repeat(x, t):
    if not isinstance(x, list):
        return list(repeat(x, t))
    elif len(x) == 1:
        return list(repeat(x[0], t))
    else:
        return x

def conv1x1(in_planes: int, out_planes: int, stride: int = 1, bias: bool = False) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=bias)

def sri_convkxk(k: int, in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1
               , kernel_shape: str = 'o', train_index_mat: bool = False, bias: bool = False, ri_k: int = None,
               force_circular: bool = False) -> SRI_Conv2d:
    """3x3 RI convolution with padding"""
    return SRI_Conv2d(in_planes, out_planes, kernel_size=k, stride=stride,
                     padding=int(k//2), groups=groups, bias=bias, dilation=dilation,
                     kernel_shape=kernel_shape, train_index_mat=train_index_mat,
                     ri_k=ri_k, force_circular=force_circular)



class SRIBasicBlock(nn.Module):
    """
    Avoid the issue due to RIconv with stride differ to 1
    """
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        kernel_shape: str = 'o',
        train_index_mat: bool = False,
        ri_conv_size: int = 3,
        ri_groups: int = 1,
        ri_k: int = None,
        force_circular: bool = False,
        **kwargs: Any,
    ) -> None:
        super(SRIBasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.stride = stride
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        conv1 = sri_convkxk(ri_conv_size, inplanes, planes, 
                            kernel_shape=kernel_shape, 
                            train_index_mat=train_index_mat, 
                            groups=ri_groups, ri_k=ri_k, 
                            force_circular=force_circular)
        if self.stride != 1:
            self.conv1 = nn.Sequential(
                nn.AvgPool2d(kernel_size=self.stride, stride=self.stride),
                conv1,
            )
        else:
            self.conv1 = conv1
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = sri_convkxk(ri_conv_size, planes, planes, kernel_shape=kernel_shape, 
                                train_index_mat=train_index_mat, groups=ri_groups,
                                ri_k=ri_k)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        
    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class SRIBottleneck(nn.Module):
    """
    Symmetric Residual Bottleneck Block
    """

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        kernel_shape: str = 'o',
        train_index_mat: bool = False,
        ri_conv_size: int = 3,
        ri_groups: int = 1,
        ri_index_mat_cin: bool = False,
        ri_index_mat_cout: bool = False,
        ri_k: int = None,
        force_circular: bool = False,
        **kwargs: Any,
    ) -> None:
        super(SRIBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        ri_groups = width if ri_groups != 1 else 1
        self.stride = stride
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        # conv1 and conv3 is the same in all case
        conv2 = sri_convkxk(ri_conv_size, width, width, stride=1, kernel_shape=kernel_shape, 
                            train_index_mat=train_index_mat, groups=ri_groups,
                            ri_index_mat_cin=ri_index_mat_cin, ri_index_mat_cout=ri_index_mat_cout,
                            ri_k=ri_k, force_circular=force_circular)
        if self.stride != 1:
            self.conv2 = nn.Sequential(
                nn.AvgPool2d(kernel_size=self.stride, stride=self.stride),
                conv2
            )
        else:
            self.conv2 = conv2
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class SRI_ResNet(nn.Module):
    """
    Here we only introduce new parameters for SRI_ResNet.
    
    Args:
        large_conv: if True, use 5x5 conv in the first layer, otherwise use 3x3 conv.
        kernel_shape: shape of the SRI convolutional kernel. default: 'o'
        train_index_mat: if True, train the index matrix. default: False
        ri_conv_size: size of the SRI convolutional kernel, can be a list of length 4. default: 3
        deepwise_ri: if True, use deepwise SRI convolution. default: False
        inplanes: number of base channels. default: 64
        layer_stride: per-layer stride for each stage in the ResNet. default: [1, 2, 2, 2]
        ri_k: number of bands in the index matrix. default: None
        force_circle: if True, force the SRI weight matrix to be circular. default: False
        skip_first_maxpool: if True, skip the first maxpool layer, set True when processing small images. default: False
    """

    def __init__(
        self,
        block: Type[Union[SRIBasicBlock, SRIBottleneck]],
        layers: List[int],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        large_conv: bool = False,
        kernel_shape: str = 'o',
        train_index_mat: bool = False,
        ri_conv_size: Union[int, list] = 3,
        deepwise_ri: bool = False,
        inplanes: int = 64,
        layer_stride: Type[Union[int, List[int]]] = [1, 2, 2, 2],
        ri_k: Union[int, list] = None,
        force_circular: bool = False,
        skip_first_maxpool: bool = False,
        **kwargs,
    ) -> None:
        super(SRI_ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.large_conv = large_conv
        self.kernel_shape = kernel_shape
        self.train_index_mat = train_index_mat
        self.ri_conv_size = _repeat(ri_conv_size, 4)
        self.deepwise_ri = deepwise_ri
        self.ri_k = _repeat(ri_k, 4)
        self.force_circular = force_circular
        if isinstance(layer_stride, float):
            layer_stride = [layer_stride for _ in range(4)]
        else:
            layer_stride = layer_stride

        self.inplanes = inplanes
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group

        self.conv1 = SRI_Conv2d(3, self.inplanes, kernel_size=5, stride=1, 
                                padding=2, bias=False, kernel_shape=kernel_shape, 
                                train_index_mat=train_index_mat, 
                                force_circular=force_circular)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        if skip_first_maxpool:
            self.maxpool = nn.Identity()
        else:
            self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, inplanes, layers[0], stride=layer_stride[0],
                                       ri_k=self.ri_k[0], ri_conv_size=self.ri_conv_size[0])
        self.layer2 = self._make_layer(block, 2*inplanes, layers[1], stride=layer_stride[1],
                                       dilate=replace_stride_with_dilation[0],
                                       ri_k=self.ri_k[1], ri_conv_size=self.ri_conv_size[1])
        self.layer3 = self._make_layer(block, 4*inplanes, layers[2], stride=layer_stride[2],
                                       dilate=replace_stride_with_dilation[1],
                                       ri_k=self.ri_k[2], ri_conv_size=self.ri_conv_size[2])
        self.layer4 = self._make_layer(block, 8*inplanes, layers[3], stride=layer_stride[3],
                                       dilate=replace_stride_with_dilation[2],
                                       ri_k=self.ri_k[3], ri_conv_size=self.ri_conv_size[3])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.flatten = nn.Flatten(1)
        self.fc = nn.Linear(8 * inplanes * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, SRI_Conv2d):
                _, fan_out = nn.init._calculate_fan_in_and_fan_out(
                    torch.zeros(m.weight_matrix_shape)
                )
                gain = nn.init.calculate_gain('relu', 0)
                std = gain / math.sqrt(fan_out)
                nn.init.normal_(m.weight, 0, std)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, SRIBottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, SRIBasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: Type[Union[SRIBasicBlock, SRIBottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False, ri_conv_size: int = 3, ri_k: int = None) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.AvgPool2d(stride, stride),
                conv1x1(self.inplanes, planes * block.expansion),
                norm_layer(planes * block.expansion),
            )

        layers = []
        ri_groups = planes if self.deepwise_ri else 1
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer, 
                            kernel_shape=self.kernel_shape, train_index_mat=self.train_index_mat,
                            ri_conv_size=ri_conv_size, ri_groups=ri_groups,
                            large_conv=self.large_conv, ri_k=ri_k, force_circular=self.force_circular))
        self.inplanes = planes * block.expansion
        ri_groups = planes if self.deepwise_ri else 1
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer, kernel_shape=self.kernel_shape, 
                                train_index_mat=self.train_index_mat,
                                ri_conv_size=ri_conv_size, ri_groups=ri_groups,
                                large_conv=self.large_conv,
 ri_k=ri_k, force_circular=self.force_circular))

        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = self.flatten(x)
        x = self.fc(x)

        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def _resnet(
    arch: str,
    block: Type[Union[SRIBasicBlock, SRIBottleneck]],
    layers: List[int],
    **kwargs: Any
) -> SRI_ResNet:
    model = SRI_ResNet(block, layers, **kwargs)
    return model


def sri_resnet18(**kwargs: Any) -> SRI_ResNet:
    r"""SRI_ResNet-18 model from
    'SRI-Conv: Symmetric Rotation-Invariant Convolutional Kernel'
    """
    return _resnet('sri_resnet18', SRIBasicBlock, [2, 2, 2, 2], **kwargs)

def sri_resnet50(**kwargs: Any) -> SRI_ResNet:
    r"""SRI_ResNet-50 model from
    'SRI-Conv: Symmetric Rotation-Invariant Convolutional Kernel'
    """
    return _resnet('sri_resnet50', SRIBottleneck, [3, 4, 6, 3], **kwargs)


import math
import collections
from itertools import repeat
from typing import List, Union, Tuple, Optional

from scipy import ndimage
import numpy as np
import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F
from torch.nn.common_types import _size_2_t

def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return tuple(x)
        return tuple(repeat(x, n))

    return parse

_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)

def _reverse_repeat_tuple(t, n):
    r"""Reverse the order of `t` and repeat each element for `n` times.

    This can be used to translate padding arg used by Conv and Pooling modules
    to the ones used by `F.pad`.
    """
    return tuple(x for x in reversed(t) for _ in range(n))


class SRI_Conv2d(nn.Module):
    
    r"""
    Official implementation of Symmetric Rotation Invariant Conv2d in PyTorch.
    Apply symmetric rotation invariant convolution over an input signal composed of several input planes.
    
    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0
        dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
        transposed (bool, optional): If ``True``, use a transposed convolution operator. Default: ``False``
        groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the output. Default: ``True``
        padding_mode (str, optional): Accepted values `zeros`, `reflect`, `replicate`, `circular`. Default: `zeros`
        kernel_shape (str, optional): Shape of constructed kernel. Accepted values `o`, `s`. Default: `o`
        ri_k (int, optional): Number of bands in the SRI kernel. If None, calculate based on kernel size. Default: None
        train_index_mat (bool, optional): If ``True``, train the index matrix. Default: ``False``
        inference_accelerate (bool, optional): If ``True``, accelerate inference by pre-computing the weight matrix. Default: ``True``
        force_circular (bool, optional): If ``True``, force the kernel to be circular by ignoring corner part. Default: ``False``
        device (torch.device, optional): Device to use. Default: ``None``
        dtype (torch.dtype, optional): Data type to use. Default: ``None``
    
    Attributes:
        weight (torch.Tensor): the learnable weights of the module of shape
            :math:`(\text{out\_channels}, \frac{\text{in\_channels}}{\text{groups}}, \text{ri\_k})`
        bias (torch.Tensor):   the learnable bias of the module of shape
            :math:`(\text{out\_channels})`
        weight_index_mat (torch.Tensor): the learnable index matrix of the module of shape
            :math:`(\text{in\_channels}, \text{out\_channels}, \text{ri\_k}, \text{ri\_k})`
        infer_weight_matrix (torch.Tensor): the pre-computed weight matrix of the module of shape
            :math:`(\text{out\_channels}, \frac{\text{in\_channels}}{\text{groups}}, \text{H}, \text{W})`
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        transposed: bool = False,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = 'zeros',
        kernel_shape: str = 'o', # ['o', 's']
        ri_k: int = None,
        train_index_mat: bool = False,
        inference_accelerate: bool = True,
        force_circular: bool = False,
        device = None,
        dtype = None,
    ) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        dilation = _pair(dilation)
        padding = padding if isinstance(padding, str) else _pair(padding)
        super().__init__()
        if in_channels % groups != 0:
            raise ValueError('in_channels must be divisible by groups')
        if out_channels % groups != 0:
            raise ValueError('out_channels must be divisible by groups')
        valid_padding_strings = {'same', 'valid'}
        if isinstance(padding, str):
            if padding not in valid_padding_strings:
                raise ValueError(
                    "Invalid padding string {!r}, should be one of {}".format(
                        padding, valid_padding_strings))
            if padding == 'same' and any(s != 1 for s in stride):
                raise ValueError("padding='same' is not supported for strided convolutions")
        valid_padding_modes = {'zeros', 'reflect', 'replicate', 'circular'}
        if padding_mode not in valid_padding_modes:
            raise ValueError("padding_mode must be one of {}, but got padding_mode='{}'".format(
                valid_padding_modes, padding_mode))
        valid_kernel_shape = {'o', 's'}
        if kernel_shape not in valid_kernel_shape:
            raise ValueError("kernel_shape must be one of {}, but got kernel_shape='{}'".format(
                valid_kernel_shape, kernel_shape))

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.transposed = transposed
        self.groups = groups
        self.padding_mode = padding_mode
        self.kernel_shape = kernel_shape
        self.output_padding = self.padding
        self.train_index_mat = train_index_mat
        self.inference_accelerate = inference_accelerate
        self.force_circular = force_circular

        self.ri_k = (kernel_size[0] // 2) + 1
        if self.kernel_shape == 'o' and not self.force_circular:
            self.ri_k += 1
        self.ri_k = self.ri_k if ri_k == None else ri_k

        if isinstance(self.padding, str):
            self._reversed_padding_repeated_twice = [0, 0] * len(kernel_size)
            if padding == 'same':
                for d, k, i in zip(dilation, kernel_size,
                                   range(len(kernel_size) - 1, -1, -1)):
                    total_padding = d * (k - 1)
                    left_pad = total_padding // 2
                    self._reversed_padding_repeated_twice[2 * i] = left_pad
                    self._reversed_padding_repeated_twice[2 * i + 1] = (
                        total_padding - left_pad)
        else:
            self._reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)

        if transposed:
            self.weight = nn.Parameter(torch.empty(
                (in_channels, out_channels // groups, self.ri_k), **factory_kwargs))
            self.weight_matrix_shape = (in_channels, out_channels // groups, *kernel_size)
            weight_index_mat = self._make_weight_index_mat(self.kernel_shape, 1, 
                                                           1, factory_kwargs)
        else:
            self.weight = nn.Parameter(torch.empty(
                (out_channels, in_channels // groups, self.ri_k), **factory_kwargs))
            self.weight_matrix_shape = (out_channels, in_channels // groups, *kernel_size)
            weight_index_mat = self._make_weight_index_mat(self.kernel_shape, 1, 
                                                           1, factory_kwargs)
        if bias:
            self.bias = nn.Parameter(torch.empty(out_channels, **factory_kwargs))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

        # use to clip the weight index mat
        self.register_buffer('default_weight_index_mat', weight_index_mat)
        if self.train_index_mat:
            self.weight_index_mat = nn.Parameter(weight_index_mat)
            self.weight_index_mat.requires_grad = True
        else:
            self.register_buffer('weight_index_mat', weight_index_mat)

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(k), 1/sqrt(k)), where k = weight.size(1) * prod(*kernel_size)
        # For more details see: https://github.com/pytorch/pytorch/issues/15314#issuecomment-477448573
        # manually init weight to avoid problem
        fan, _ = init._calculate_fan_in_and_fan_out(torch.zeros(self.weight_matrix_shape))
        gain = init.calculate_gain('leaky_relu', math.sqrt(5))
        std = gain / math.sqrt(fan)
        bound = math.sqrt(3.0) * std
        init.uniform_(self.weight, -bound, bound)
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def extra_repr(self):
        s = ('{in_channels}, {out_channels}, kernel_size={kernel_size}'
             ', stride={stride}')
        if self.padding != (0,) * len(self.padding):
            s += ', padding={padding}'
        if self.dilation != (1,) * len(self.dilation):
            s += ', dilation={dilation}'
        if self.output_padding != (0,) * len(self.output_padding):
            s += ', output_padding={output_padding}'
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        if self.padding_mode != 'zeros':
            s += ', padding_mode={padding_mode}'
        return s.format(**self.__dict__)

    def __setstate__(self, state):
        super(SRI_Conv2d, self).__setstate__(state)
        if not hasattr(self, 'padding_mode'):
            self.padding_mode = 'zeros'

    def _make_weight_index_mat(
        self, kernel_shape, 
        index_mat_C_in, 
        index_mat_C_out,
        factory_kwargs
    ):
        weight_index_mats = []
        _, _, H, W = self.weight_matrix_shape
        if kernel_shape == 's':
            for i in range(self.ri_k):
                mat = torch.zeros([H, W], **factory_kwargs)
                if i == self.ri_k - 1:
                    mat[i, i] = 1
                else:
                    mat[i, i:2*self.ri_k-1-i] = 1
                    mat[2*self.ri_k-2-i, i:2*self.ri_k-1-i] = 1
                    mat[i+1:2*self.ri_k-2-i, i] = 1
                    mat[i+1:2*self.ri_k-i-2, 2*self.ri_k-2-i] = 1
                weight_index_mats.append(mat.reshape(-1))
        elif kernel_shape == 'o':
            D = np.ones((H, W))
            D[(H//2, W//2)] = 0
            D = ndimage.distance_transform_edt(D)
            max_dist = (H // 2) + 0.5 if self.force_circular else D.max()
            num_levels = self.ri_k + 1 if self.force_circular else self.ri_k
            levels = np.linspace(D.min(), max_dist, num=num_levels)
            for i in range(num_levels):
                if i == num_levels - 1:
                    if self.force_circular:
                        continue
                    idx = (D==levels[i]).astype(int)
                else:
                    idx = ((D>=levels[i]) & (D<levels[i+1])).astype(int)
                level_mat = torch.tensor(idx, **factory_kwargs)[None, None, :, :]
                weight_index_mats.append(level_mat.reshape(-1))
        weight_index_mats = torch.stack(weight_index_mats, dim=0)
        weight_index_mats = weight_index_mats.to(torch.float32)
        weight_index_mats = weight_index_mats.expand((index_mat_C_in, index_mat_C_out, -1, -1))
        return weight_index_mats


    def _make_weight_matrix(self, weight):
        # Note that einsum is generally faster than batch matrix multiplication 
        weight = torch.einsum('ijkw,ijk->ijw', self.weight_index_mat, weight)
        weight = weight.reshape(self.weight_matrix_shape)
        return weight


    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            return F.conv2d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                            weight, bias, self.stride, _pair(0), self.dilation, self.groups)
        return F.conv2d(input, weight, bias, self.stride,
                        self.padding, self.dilation, self.groups)


    def forward(self, input):
        if not self.training and self.inference_accelerate:
            weight_matrix = self.infer_weight_matrix
        else:
            weight_matrix = self._make_weight_matrix(self.weight)
        return self._conv_forward(input, weight_matrix, self.bias)


    def clip_index_mat(self, eps=0.1):
        if not self.train_index_mat:
            return
        # only clip the index mat after each param update
        _max = self.default_weight_index_mat + eps
        _min = self.default_weight_index_mat - eps
        self.weight_index_mat.data = torch.clamp(self.weight_index_mat, min=_min, max=_max)


    def train(self, mode=True):
        if not isinstance(mode, bool):
            raise ValueError("training mode is expected to be boolean")
        if self.inference_accelerate and not mode:
            self.infer_weight_matrix = self._make_weight_matrix(self.weight)
        else:
            self.infer_weight_matrix = None
        return super().train(mode)


class SRI_ConvTranspose2d(SRI_Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode='zeros',
        device=None,
        dtype=None,
        kernel_shape='s',
        train_index_mat=False,
        ri_k=None,
        index_mat_in_channels=False,
        index_mat_out_channels=False
    ) -> None:
        super().__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, True, groups, bias, padding_mode, device, dtype, kernel_shape, train_index_mat, ri_k, index_mat_in_channels, index_mat_out_channels)


    # dilation being an optional parameter is for backwards
    # compatibility
    def _output_padding(self, input, output_size,
                        stride, padding, kernel_size,
                        num_spatial_dims, dilation = None):
        if output_size is None:
            ret = _single(self.output_padding)  # converting to list if was not already
        else:
            has_batch_dim = input.dim() == num_spatial_dims + 2
            num_non_spatial_dims = 2 if has_batch_dim else 1
            if len(output_size) == num_non_spatial_dims + num_spatial_dims:
                output_size = output_size[num_non_spatial_dims:]
            if len(output_size) != num_spatial_dims:
                raise ValueError(
                    "ConvTranspose{}D: for {}D input, output_size must have {} or {} elements (got {})"
                    .format(num_spatial_dims, input.dim(), num_spatial_dims,
                            num_non_spatial_dims + num_spatial_dims, len(output_size)))

            min_sizes = torch.jit.annotate(List[int], [])
            max_sizes = torch.jit.annotate(List[int], [])
            for d in range(num_spatial_dims):
                dim_size = ((input.size(d + num_non_spatial_dims) - 1) * stride[d] -
                            2 * padding[d] +
                            (dilation[d] if dilation is not None else 1) * (kernel_size[d] - 1) + 1)
                min_sizes.append(dim_size)
                max_sizes.append(min_sizes[d] + stride[d] - 1)

            for i in range(len(output_size)):
                size = output_size[i]
                min_size = min_sizes[i]
                max_size = max_sizes[i]
                if size < min_size or size > max_size:
                    raise ValueError((
                        "requested an output size of {}, but valid sizes range "
                        "from {} to {} (for an input of {})").format(
                            output_size, min_sizes, max_sizes, input.size()[2:]))

            res = torch.jit.annotate(List[int], [])
            for d in range(num_spatial_dims):
                res.append(output_size[d] - min_sizes[d])

            ret = res
        return ret


    def forward(self, input, output_size = None):
        if self.padding_mode != 'zeros':
            raise ValueError('Only `zeros` padding mode is supported for ConvTranspose2d')

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 2
        output_padding = self._output_padding(
            input, output_size, self.stride, self.padding, self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims, self.dilation)  # type: ignore[arg-type]

        weight_matrix = self._make_weight_matrix(self.weight)

        return F.conv_transpose2d(
            input, weight_matrix, self.bias, self.stride, self.padding,
            output_padding, self.groups, self.dilation)

